﻿#include "emulator/Emulator.hpp"

#include <conio.h>
#include <windows.h>

#include <iomanip>
#include <iostream>
#include <sstream>
#include <stdexcept>

class Application {
	static SHORT round(double v) //
	{
		return static_cast<SHORT>(std::round(v));
	}

	void print(Location location, char c, Car& car)
	{
		// Напечатать символ на координатах

		const auto x = round(location.x - car.getLocation().x) + 30;
		const auto y = round(location.y - car.getLocation().y) + 10;

		if (x > 10 && y > 6 && x < 90 && y < 20)
		{
			COORD pos = { x, y };
			SetConsoleCursorPosition(mOutput, pos);
			std::cout << c;
		}
	}

	void print(Location location, char c)
	{
		// Напечатать символ на координатах

		const auto x = round(location.x);
		const auto y = round(location.y);

		//if (x > 0 && y > 0) {
		COORD pos = { x, y };
		SetConsoleCursorPosition(mOutput, pos);
		std::cout << c;
		//}
	}

	int getchar()
	{
		// Читаем один символ если такие там есть
		if (_kbhit())
			return _getch();
		return 0;
	}

public:
	explicit Application(Mode mode)
		: mMode(mode)
		, mOutput(GetStdHandle(STD_OUTPUT_HANDLE))
	{
		if (mMode == Mode::HandsOff) {
			// Режим автоматического управления
		}
		if (mMode == Mode::HandsOn) {
			// Режим ручного управления
		}
	}

	bool run(State& state)
	{
		bool result = true;

		// Выходим, потому что не умеем управлять вручную
		if (mMode == Mode::HandsOn)
		{
			result = handsMode(state);
		}
		else
		{
			result = handsOffMode(state);
		}

		// Пробуем что-то отобразить
		draw(state);

		// Продолжаем работу если мы не достигли точки назначения или не столкнулись
		return !state.isDestination() && !state.isCollision() && result;
	}

	void draw(State& state)
	{
		auto& car = state.getCar();

		system("cls");

		for (const auto& lane : state.getRoad().getLanes())
		{
			auto startX = lane.getLocation().x;
			auto endX = startX + lane.getDirection().dx;
			auto pieceOfRoad = lane.getLocation();

			for (int i = startX; i < endX; i++)
			{
				pieceOfRoad.x = i;
				auto startY = lane.getLocation().y;
				auto endY = startY + lane.getDirection().dy;

				for (int j = startY; j < endY; j++)
				{
					pieceOfRoad.y = j;

					mRoads.push_back(pieceOfRoad);
					print(pieceOfRoad, '.', car);
				}

			}
		}

		for (auto& sign : state.getRoad().getSigns())
		{
			auto lane = state.getRoad().getLanes()[sign.getLaneIndex()];

			auto startX = sign.getLocation();
			auto startY = lane.getLocation().y;
			auto endY = lane.getDirection().dy + startY;

			auto location = Location
			{
				startX,
				startY
			};

			for (int i = startY; i < endY; i++)
			{
				location.y = i;

				mLocations.push_back(location);
				print(location, '*', car);
			}
		}

		// Добавляем символ автомобиля
		{
			const auto location = car.getLocation();
			mLocations.push_back(location);
			print(location, '@', car);
		}

		// Добавляем остальные автомобили
		for (const auto& vehicle : state.getVisibleVehicles())
		{
			const auto location = vehicle.getLocation();
			mLocations.push_back(location);
			print(location, 'o', car);
		}

		// Рисуем финиш
		{
			auto finsih = car.getDestination();

			auto radius = finsih.radius;
			auto startX = finsih.center.x - radius;
			auto endX = finsih.center.x + radius;
			auto startY = finsih.center.y - radius;
			auto endY = finsih.center.y + radius;

			auto finishPiece = Location();
			finishPiece.y = finsih.center.y;

			for (int x = startX; x < endX; x++)
			{
				finishPiece.x = x;
				print(finishPiece, '#', car);
			}

			finishPiece.x = finsih.center.x;
			for (int y = startY; y < endY; y++)
			{
				finishPiece.y = y;
				print(finishPiece, '#', car);
			}
		}

		{
			auto speedLocation = Location{ 0, 0 };

			auto speed = car.getCurrentSpeed();
			auto speedStr = "car speed: " + std::to_string(speed);
			printStringAtLocation(speedStr, speedLocation);

			auto gearLocation = Location{ 30, 0 };
			printStringAtLocation(mGear == Gear::Forward ? "F" : mGear == Gear::Neutral ? "N" : "R", gearLocation);

			auto location = Location{ 0, 1 };

			printStringAtLocation("@ - your car, o - another cars, * - traffic lights, # - finish", location);

			auto carPos = car.getLocation();
			auto destanation = car.getDestination();

			location.x = 0;
			location.y = 2;

			auto lengthToDestenation = getLengthByPoints(carPos.x, carPos.y, destanation.center.x, destanation.center.y);

			auto lengthStr = std::to_string(lengthToDestenation);

			printStringAtLocation("Length to destatnation: " + lengthStr, location);
		}
	}

	void printStringAtLocation(std::string str, Location startLocation)
	{
		for (auto& element : str)
		{
			mLocations.push_back(startLocation);
			print(startLocation, element);
			startLocation.x++;
		}
	}

private:
	HANDLE mOutput;
	Mode mMode;
	Gear mGear = Gear::Neutral;
	std::vector<Location> mLocations;
	std::vector<Location> mRoads;
	AngleUnit mRotateValue;
	LengthUnit speedRule = 1000;

	/// <summary>
	/// Ручной режим перемещения
	/// </summary>
	/// <param name="state"></param>
	/// <returns></returns>
	bool handsMode(State& state)
	{
		bool result = true;
		auto pressedKey = tolower(getchar());

		Car& car = state.getCar();

		if (pressedKey == 'w' || pressedKey == 230)
		{
			if (mGear == Gear::Reverse)
			{
				car.setGear(Gear::Neutral);
				mGear = Gear::Neutral;
			}
			else
			{
				car.setGear(Gear::Forward);
				mGear = Gear::Forward;
			}
		}
		else if (pressedKey == 's' || pressedKey == 235)
		{
			if (mGear == Gear::Forward)
			{
				car.setGear(Gear::Neutral);
				mGear = Gear::Neutral;
			}
			else
			{
				car.setGear(Gear::Reverse);
				mGear = Gear::Reverse;
			}
		}
		else if (pressedKey == 'a' || pressedKey == 228)
		{
			if (mRotateValue > 0)
			{
				mRotateValue = 0;
				car.turnSteeringWheel(0);
			}
			else
			{
				mRotateValue = -0.25;
				car.turnSteeringWheel(-0.25);
			}
		}
		else if (pressedKey == 'd' || pressedKey == 162)
		{
			if (mRotateValue < 0)
			{
				mRotateValue = 0;
				car.turnSteeringWheel(0);
			}
			else
			{
				mRotateValue = 0.25;
				car.turnSteeringWheel(0.25);
			}
		}
		else if (pressedKey == 'e' || pressedKey == 227)
		{
			car.pressGasPedal(0.2);
		}
		else if (pressedKey == 'q' || pressedKey == 169)
		{
			car.pressGasPedal(-0.2);
		}

		return result;
	}

	/// <summary>
	/// Автоматический режим перемещения
	/// </summary>
	/// <param name="state"></param>
	/// <returns></returns>
	bool handsOffMode(State& state)
	{
		// Мой автомобиль
		auto& car = state.getCar();

		auto destanation = car.getDestination();
		auto carPos = car.getLocation();

		auto cSide = abs(carPos.x - destanation.center.x);
		auto bSide = abs(carPos.y - destanation.center.y);
		auto aSide = sqrt(bSide* bSide + cSide * cSide);

		auto angleToFinsih = acos((bSide * bSide + cSide * cSide - aSide * aSide) / (2 * bSide * cSide)) * 180 / 3.14;

		car.getLocation();

		auto lengthToDestenation = getLengthByPoints(carPos.x, carPos.y, destanation.center.x, destanation.center.y);
		
		double nearestTraffic = 100;

		for (auto& trafficSign : state.getRoad().getSigns())
		{
			auto& lane = state.getRoad().getLanes()[trafficSign.getLaneIndex()];

			if (nearestTraffic > (lane.getLocation().x + trafficSign.getLocation()))
			{
				nearestTraffic = lane.getLocation().x + trafficSign.getLocation() - car.getLocation().x;

				if (nearestTraffic < 0)
				{
					speedRule = trafficSign.getValue();
				}
			}
		}

		if (car.getCurrentSpeed() < car.getMaxSpeed(Gear::Forward) * 0.8 && speedRule == 1000)
		{
			car.setGear(Gear::Forward);
			car.pressGasPedal(car.getMaxSpeed(Gear::Forward) * 0.8);
		}
		else if (car.getCurrentSpeed() > speedRule)
		{
			car.pressBrakePedal(car.getMaxSpeed(Gear::Forward));
		}
		else if (car.getCurrentSpeed() < speedRule)
		{
			car.setGear(Gear::Forward);
			car.pressGasPedal(car.getMaxSpeed(Gear::Forward) * 0.8);
		}
		else
		{
			car.setGear(Gear::Neutral);
		}

		if (lengthToDestenation < (car.getDestination().radius*2))
		{
			car.pressBrakePedal(car.getMaxSpeed(Gear::Forward));
		}

		return true;
	}

	bool findInRoads(Location location)
	{
		for (auto& road : mRoads)
		{
			if (compareDoubles(road.x, location.x) && compareDoubles(road.y, location.y))
			{
				return true;
			}
		}

		return false;
	}

	void removeFromRoadsIfExists(Location location)
	{
		int i = 0;
		for (auto& road : mRoads)
		{
			if (compareDoubles(road.x, location.x) && compareDoubles(road.y, location.y))
			{
				mRoads.erase(mRoads.begin() + i);
			}
			i++;
		}
	}

	bool compareDoubles(double first, double second)
	{
		return std::fabs(first - second) < std::numeric_limits<double>::epsilon();
	}

	double getLengthByPoints(double x1, double y1, double x2, double y2)
	{
		auto sum = pow(x2 - x1, 2) + pow(y2 - y1, 2);
		return sqrt(sum);
	}
};

int main(int argc, char** argv)
{
	try
	{
		Emulator emulator(argc, argv);

		Application app(emulator.getMode());

		emulator.run([&app](State& state) { return app.run(state); });

		return 0;
	}
	catch (const std::runtime_error& error)
	{
		std::cerr << error.what() << std::endl;
		return 1;
	}
}
